/*
   rads by Aaron Hachez. 

   All code copyright (C) 2004, Aaron Hachez.

   rads is a utility that allows one to parse records from text files.

   ----------------------------------------------------------------------------------------

   This file is part of "rads".

   A copy of the GNU General Public License should be included with any
   redistribution of this program in a file named "COPYING.txt".

   "rads" is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   "rads" is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with "rads"; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   ----------------------------------------------------------------------------------------
*/

#include <cstdio>
#include <cstring>
#include <cstdlib>

// The version
#define VERSION "v1.8.5"

// The max size for the delimiter
#define DELIMLEN 256
#define SEARCHLEN 256

// The difference between cap characters
#define CAPDIFF ('a' - 'A')

int printver()
{
    printf("rads %s\n\n", VERSION);
    printf("Copyright (C) 2001, 2004, Aaron Hachez\n");
    printf("This is free software; see the source for copying conditions. There is NO\n");
    printf("warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");

    return(0);
}

int printusage()
{
    printf("Usage: rads [-hiIlnqrtV] [--d=<delimiter>] <searchstring> <file1> <file2> <etc...>\n");

    return(0);
}

int printhelp()
{
    printusage();

    printf("\n -h                 Print this help.\n");
    printf(" -i                 Perform case insensitive search.\n");
    printf(" -I                 Perform case insensitive search for delimiters.\n");
    printf(" -r                 Turn on record separators.\n");
    printf(" -q                 Don't be verbose.\n");
    printf(" -l                 Do not print leading delimiter.\n");
    printf(" -t                 Do not print trailing delimiter.\n");
    printf(" -n                 Do not print leading or trailing delimiter.\n");
    printf(" -V                 Print version information and exit.\n\n");
    printf(" --d=<delimiter>    Set the delimiter for records (Default is '\\n\\n')\n");
    printf("\n(All single dash (-) options may be combined)\n\n");
    printf("Certain special characters may be specified in both delimiter and search \n");
    printf("fields, and must be escaped. Supported characters are:\n\n"); 
    printf("  Carriage Return -- \\r\n");
    printf("  Line Feed       -- \\n\n");
    printf("  Tab             -- \\t\n");
    printf("\nIn addition, if you wish to use \\ it too must be escaped (\\\\).\n");
    printf("\nDepending upon your shell environment, you may also have to escape\n");
    printf("your escapes -- i.e. \\\\n, \\\\t, \\\\ -- and enclose your term in quotes.\n\n");
    printf("To search using two line feeds one tab and the letter h as your delimiter:\n\n");
    printf(" rads --d=\"\\\\n\\\\n\\\\th\" <searchterm> <file1> <file2> <etc...>\n\n");

    return(0);
}

// Match two characters, caseins determines whether they should be matched according to case or not.
// Returns 1 on match, 0 on no match
inline
int charmatch(char a, char b, int caseins)
{
    int result = 0;

    // Be case insensitive
    if(caseins > 0)
    {
        // Here we lowercase the char, but only if it is a cap char.
        if((a >= 'A') && (a <= 'Z'))
        {
            a = a + CAPDIFF;
        }

        // Here we lowercase the char, but only if it is a cap char.
        if((b >= 'A') && (b <= 'Z'))
        {
            b = b + CAPDIFF;
        }
    }

    if(a == b)                                  // If we have a match
    {
        result = 1;
    }
    else
    {
        result = 0;
    }

    return(result);
}

int main(int argc, char* argv[])
{

    /* 
    Check inputs
    Get search string
    Open file descripter
    Loop through file
     for char in search string
      if match next
      else reset search counter
    close file
    loop
    */

    char searchstr[SEARCHLEN];                  // The string we are searching for
    memset(searchstr, '\0', SEARCHLEN);         // Prepare memory for search string

    char delimiter[DELIMLEN];                   // The record delimiter
    memset(delimiter, '\0', DELIMLEN);

    int printdelimiterleading = 1;              // Whether or not to print the leading delimiter (On by default)
    int printdelimitertrailing = 1;             // Whether or not to print the trailing delimiter (On by default)
    
    int verboseon = 1;                          // Whether or not to print verbose information
    int srchcaseinsensitive = 0;                // Whether or not to be case insensitive
    int delimcaseinsensitive = 0;               // Whether or not to be case insensitive for delimiter

    int separatoron = 0;                        // Whether or not record separators will be printed
    char separator[64] = " ---------- RECORD SEPARATOR ---------- \0";

    int fileargs = 0;                           // Where the file arguments start in the argv list
    char filename[256];                         // The name of the current file
    memset(filename, '\0', sizeof(filename));   // Prep variable

    FILE *fp = NULL;                                   // File pointer

    for(int i = 1; i < argc; ++i)               // i = 1 initially to bypass self argument.
    {
        int arglen;                             // Used to store the argument length, different for arg type

        // The first non-dash argument is the search string, all others are filenames
        if(argv[i][0] == '-')
        {
                                                                    // Arguments that are not extended may only be one char
            if(argv[i][1] != '-')                                   // If this is not an extended argument (i.e. --arg). 
            {
                arglen = strlen(argv[i]);                           // Get the length of this argument, minus 1 for "-"

                if(arglen <= 1)
                {
                    fprintf(stderr, "Error: Invalid argument %s\n", argv[i]);
                    return(1);
                }
                
                for(int j = 1; j < arglen; ++j)                     // Start j at 1 to skip past "-" (Starts at 0)
                {
                    switch(argv[i][j])
                    {
                        case 'h':                                   // Print help and exit
                            printhelp();
                            return(0);
                            break;
                        case 'r':                                   // Turn on record separators
                            separatoron = 1;
                            break;
                        case 'q':
                            verboseon = 0;                          // Don't be verbose
                            break;
                        case 't':
                            printdelimitertrailing = 0;             // Don't print trailing delimiter
                            break;
                        case 'l':
                            printdelimiterleading = 0;              // Don't print leading delimiter
                            break;
                        case 'n':
                            printdelimitertrailing = 0;             // Don't print trailing or leading delimiter
                            printdelimiterleading = 0;
                            break;
                        case 'i':
                            srchcaseinsensitive = 1;                // Be case insensitive
                            break;
                        case 'I':
                            delimcaseinsensitive = 1;               // Be case insensitive for delimiter
                            break;
                        case 'V':
                            printver();                             // Be case insensitive for delimiter
                            return(0);
                            break;
                        default:
                            fprintf(stderr, "Error: Invalid option %c\n", argv[i][j]);
                            return(1);
                            break;
                    }
                }
            }                                   // End processing arg
            else
            {
                switch(argv[i][2])                      // This is an extended arg (i.e. with "--" rather than "-")
                {
                    case 'd':                           // Set the delimiter
                        arglen = strlen(argv[i]) - 4;   // Get the length of argv, less the argument "--d="

                        // Ensure that the argument is shorter than the length of the delimiter
                        if((arglen < DELIMLEN) && (arglen > 0))
                        {
                            memset(delimiter, '\0', DELIMLEN);
                            memcpy(delimiter, &argv[i][4], arglen);

                            // Now loop through the delimiter looking for escaped chars
                            char delimcpy[DELIMLEN];
                            memset(delimcpy, '\0', DELIMLEN);

                            int j = 0;
                            int k = 0;
                            for(j = 0; j < DELIMLEN; ++j)
                            {
                                if(delimiter[j] == '\0')        // If the char is null, exit the for loop
                                    break;

                                if(delimiter[j] == '\\')        // Two \\'s because one escapes the ', rather than matching '\'
                                {
                                    switch(delimiter[j + 1])    // Test on the next char after "\"
                                    {
                                        case 'n':
                                            delimcpy[k] = '\n'; // It was a line feed
                                            ++j;                // Jump past this char
                                            break;
                                        case 'r':
                                            delimcpy[k] = '\r'; // Cariage return
                                            ++j;                // Jump past this char
                                            break;
                                        case 't':
                                            delimcpy[k] = '\t'; // Tab
                                            ++j;                // Jump past this char
                                            break;
                                        default:
                                            delimcpy[k] = delimiter[j]; // Else copy it verbatim
                                    }
                                }
                                else
                                {
                                    delimcpy[k] = delimiter[j]; // Normal char, just copy
                                }

                                ++k;                            // Increment copy counter
                            }
                            // Make our final copy of the delimiter
                            memcpy(delimiter, delimcpy, DELIMLEN);
                        }
                        else
                        {
                            if(arglen >= DELIMLEN)
                            {
                                fprintf(stderr, "Error: The delimiter is too long, it must be less than %d characters. (Escaped characters count as two)\n", DELIMLEN);
                            }
                            else
                            {
                                fprintf(stderr, "Error: The delimiter must be at least one character in length.\n");
                            }
                            return(1);
                        }

                        break;
                    default:
                        fprintf(stderr, "Error: Invalid argument %s\n", argv[i]);
                        return(1);

                }  // End swtich
            }      // End processing extended arg
        }
        else
        {
            if(searchstr[0] == '\0')    // If the first char of the search string is '\0' we can assume that the search string hasn't been found yet.
            {
                int stringlen = 0;
                stringlen = strlen(argv[i]);

                if((stringlen < SEARCHLEN) && (stringlen > 0))
                {
                    memset(searchstr, '\0', SEARCHLEN);     // We already did this, but better safe than sorry
                    memcpy(searchstr, argv[i], stringlen);  // Copy search string

                    // Now fix any escaped characters
                    char searchcpy[SEARCHLEN];
                    memset(searchcpy, '\0', SEARCHLEN);

                    int j = 0;
                    int k = 0;
                    for(j = 0; j < SEARCHLEN; ++j)
                    {
                        if(searchstr[j] == '\0')        // If the char is null, exit the for loop
                            break;

                        if(searchstr[j] == '\\')        // Two \\'s because one escapes the ', rather than matching '\'
                        {
                            switch(searchstr[j + 1])    // Test on the next char after "\"
                            {
                                case 'n':
                                    searchcpy[k] = '\n'; // It was a line feed
                                    ++j;                // Jump past this char
                                    break;
                                case 'r':
                                    searchcpy[k] = '\r'; // Cariage return
                                    ++j;                // Jump past this char
                                    break;
                                case 't':
                                    searchcpy[k] = '\t'; // Tab
                                    ++j;                // Jump past this char
                                    break;
                                default:
                                    searchcpy[k] = searchstr[j]; // Else copy it verbatim
                            }
                        }
                        else
                        {
                            searchcpy[k] = searchstr[j]; // Normal char, just copy
                        }

                        ++k;                            // Increment copy counter
                    }
                    // Make our final copy of the searchstr
                    memcpy(searchstr, searchcpy, SEARCHLEN);
                }
                else
                {
                    fprintf(stderr, "Error: The search string must be less than %d characters, and at least one in length.\n", SEARCHLEN);
                }
            }
            else if(fileargs == 0)
            {
                fileargs = i;
            }
        }
    }

    if(delimiter[0] == '\0')                // If no delimiter passed on command line
    {
        delimiter[0] = '\n';                // Set default delimiter
        delimiter[1] = '\n';                // Set default delimiter
        delimiter[2] = '\0';                // Set default delimiter
    }

    // Okay! We are done parsing command line args, check to see if we have a search string to work with
    if(searchstr[0] == '\0')
    {
        fprintf(stderr, "Error: You must specify a search string.\n");
        printusage();
        return(0);
    }

    if(fileargs == 0)
    {
        fprintf(stderr, "Error: No file arguments.\n");
        printusage();
        return(1);
    }
    else
    {
        if(strlen(argv[fileargs]) > 255)
        {
            fprintf(stderr, "Error: filename is longer than 255 characters.\n");
            return(1);
        }

        strncpy(filename, argv[fileargs], sizeof(filename) - 1);
        fp = fopen(filename, "rb");
        ++fileargs;                             // Skip to next file
    }

    int quit = 0;
    while(quit != 1)
    {
        int records = 0;                        // The number of records found
        if(fp != NULL)
        {
            // Find and record the total length of the file
            fseek(fp, 0L, SEEK_END);
            long leof = (ftell(fp) - 1);        // The local record of the length of the file, less one for our purposes
            rewind(fp);

            char buffer[512];

            int delimlen = strlen(delimiter) - 1;   // Get length of delimiter (Str len, not total len) -- (-1 because count starts at 0, strlen starts at 1)
            int dfoundct = 0;                   // The number of characters in the delimiter that have been found
            // FIXME: starting delimiter should not always be beginning of file, this should be an option
            long dlocst = 0;                    // Where the last delimiter is (start)
            long dlocend = -1;                  // Where the last delimiter is (end)
            long dlocmaybe = -1;                // Where the next delimiter may be

            int searchlen = strlen(searchstr) - 1;  // Get length of search string (Str len, not total len) -- (-1 because count starts at 0, strlen starts at 1)
            int sfoundct = 0;                   // The number of characters in the search string that have been found
            int sfound = 0;                     // If the search string was found in toto, this is 1

            long filepos = 0;                   // Where we are in the file

            int poffsetst = 0;                  // The offset to print start delimiter or not
            int poffsetend = 0;                 // The offset to print end delimiter or not

            int match = 0;                      // Keeps track of character matches
            
            //while(!feof(fp))
            do
            {
                filepos = ftell(fp);            // Record file position before read

                memset(buffer, '\0', sizeof(buffer));       // Null out buffer
                fread(&buffer, sizeof(buffer) - 1, 1, fp);  // Read in one less than the total length of the buffer

                int wherei = 0;                             // Where the delimiter occurs
                int wheres = 0;                             // Where the search string occurs
                long realpos = 0;                           // The "real" position in file, counting buffer
                int i = 0;
                while(buffer[i] != '\0')
                {
                    realpos = filepos + i;                  // Record "real" file position
                
                    /* See if we have a match for delimiter */

                    // Here we check to see whether the current character of the delimiter and buffer match.
                    // If not, and if the number of items found from the delimiter is greater than 0, reset the counter for delimiter

                    // See if these to chars match
                    match = 0;                              // Reset match
                    match = charmatch(delimiter[dfoundct], buffer[i], delimcaseinsensitive);

                    if((dfoundct > 0) && (!match))          // Didn't match, this isn't a delimiter
                    {

                        // This re-sets i to our last known good position.
                        // Meaning that we will continue searching the buffer not where we left off, but just
                        // after where we thought the last delimiter was
                        i = wherei + 1;

                        dfoundct = 0;                       // Reset the counter
                    }

                    // FIXME: No need to do this if the above didn't succeed.
                    // Note if the above suceeded, we have to re-test the char because i is now different
                    match = 0;                              // Reset match
                    match = charmatch(delimiter[dfoundct], buffer[i], delimcaseinsensitive);

                    // Even though buffer[i] didn't match the current delimiter character, it might still be the first character in
                    // the delimiter, so we leave the following check as it's own stand alone check
                    if( match )
                    {
                        if(dfoundct == 0)                   // If this is the first char in the delimiter
                        {
                            wherei = i;                     // Record i so that if the delim doesn't match, we can reset just past this point.
                            dlocmaybe = realpos;            // Record position in file
                        }

                        if(dfoundct == delimlen)            // If we found the whole delimiter
                        {
                            if(dlocst >= 0)                 // If we already have a starting delimiter
                            {
                                dlocend = dlocmaybe;        // Set end delimiter
                            }
                            else
                            {
                                dlocst = dlocmaybe;         // Else set start delimiter
                            }

                            dfoundct = 0;
                        }
                        else
                        {
                            ++dfoundct;
                        }
                    }
                    // NOTE: What would be an else here is handled in the statment just above this one

                    /* Now see if we have a match for search string */

                    if( (dlocst >= 0) && (sfound != 1) )    // If we have the first delimiter, and the string has not yet been found
                    {
                        // Check to see if these to chars match
                        match = 0;                              // Reset match
                        match = charmatch(searchstr[sfoundct], buffer[i], srchcaseinsensitive);

                        if( match ) 
                        {
                            wheres = i;

                            if(sfoundct == searchlen)       // If the whole search string has been found
                            {
                                sfound = 1;                 // Set found flag
                                sfoundct = 0;               // Reset counter
                            }
                            else
                            {
                                ++sfoundct;                 // Else increment the counter
                            }
                        }
                        else if(sfoundct > 0)               // No match, this isn't the search string
                        {
                            sfoundct = 0;                   // Reset counter
                            i = wheres + 1;                 // Increment past the beginning of where we thought the seach string occured
                        }
                    }

                    /* If we have the start and end delimiters (Or start and eof) and a string match, no need to continue */

                    // FIXME: Make this an option, the end delimiter should not always be eof if not matched before
                    // If at the end of the file, set end delimiter to eof as stored in leof
                    if (realpos == leof)
                    {
                        // Only set to end of file, if at least one delimiter was found
                        if(dlocst != 0)
                        {
                            dlocend = leof;
                        }
                    }

                    //if( (dlocst >= 0) && ((dlocend > 0) || feof(fp)) && (sfound > 0) )
                    if( (dlocst >= 0) && (dlocend > 0) && (sfound > 0) )
                    {
                        break;                              // Note: We seek back to the start of the record, and print 
                                                            // it out, starting again at the char following record..
                    }
                    else if( (dlocst >= 0) && (dlocend > 0) && (sfound == 0) )  // If we have a full record, but no match
                    {
                        dlocst = dlocend;                   // Set start to end of record
                        dlocend = -1;                       // Clear end
                    }

                    ++i;
                }

                /* If we have start and end delimiters (Or start and eof) and the string was found */
                //if( (dlocst >= 0) && ((dlocend > 0) || feof(fp)) && (sfound > 0) )
                if( (dlocst >= 0) && (dlocend > 0) && (sfound > 0) )
                {
                    ++records;

                    if(printdelimiterleading)               // If we should print the leading delimiter
                    {
                        poffsetst = 0;                      // Start of start delim.
                    }
                    else
                    {
                        poffsetst = delimlen + 1;           // Start after start delim
                    }
               
                    if(printdelimitertrailing)
                    {
                        poffsetend = delimlen + 1;          // Stop after end of delim
                    }
                    else
                    {
                        poffsetend = 0;                     // Stop before end delim
                    }

                    fseek(fp, (dlocst + poffsetst), SEEK_SET);

                    char c;

                    if(separatoron && records > 1)
                        printf("\n%s\n", separator);        // Print separator

                    while( (ftell(fp) != (dlocend + poffsetend)) && !feof(fp) )    // While in record
                    {
                        fread(&c, sizeof(c), 1, fp);
                        printf("%c", c);
                    }

                    dlocst = dlocend;
                    dlocend = -1;
                    sfound = 0;
                }

            } while(!feof(fp));
        
            fclose(fp);
        }
        else
        {
            fprintf(stderr, "Error: Opening file %s.\n", filename);
        }

        if(verboseon)
        {
            printf("\n\n -- Found %d records in file %s -- \n", records, filename);
        }
        else
        {
            printf("\n");
        }
       
        // Open the next file
        if( (fileargs != 0) && (argv[fileargs]) )
        {
            if(strlen(argv[fileargs]) > 255)
            {
                fprintf(stderr, "Error: filename is longer than 255 characters.\n");
                return(1);
            }

            strncpy(filename, argv[fileargs], sizeof(filename) - 1);
            fp = fopen(filename, "rb");
            ++fileargs;                             // Skip to next file
        }
        else
        {
            quit = 1;
        }
    }
    return(0);
}
